
/**
 * @(#)Ball.java	1.3 99/01/13
 *
 * Ball extends Point.
 * Also maintains speed information.
 *
 * Copyright (c) 1996-1999 James Begole.  All Rights Reserved.
 *
 * James Begole MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY 
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. James Begole SHALL NOT BE LIABLE
 * FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES"). James Begole
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 *
 * @author   James "Bo" Begole
 * @author   Rich Burridge
 * @version  1.5 - 11th January 1999. 
 */

package examples.ppong;

import java.awt.*;

class 
Ball extends Point implements PpongDebugFlags {

    public static final int diameter     = 15;  // in pixels
    public static final int defaultSpeed = 90;  // in pixels/second
    
    public int XSpeed;
    public int YSpeed;
    public int originalXSpeed;
    public int originalYSpeed;

    public long xStartTime;
    public long yStartTime;

    public int startX;
    public int startY;


    public 
    Ball(int x, int y, int xspeed, int yspeed) {
        super(x, y);
        reset(x, y, xspeed, yspeed, System.currentTimeMillis());
    }


    public 
    Ball(int x, int y) {
        this(x, y, -defaultSpeed, defaultSpeed);
    }


/** true iff the ball speed should be retarded depending on the
  * network delay time.
  */

    private boolean retard = false;

/**
 * If true, retard the ball speed depending on the network lag.
 * @param value true iff ball speed should be retarded depending on the
 * network delay time
 */

    public void 
    setRetard(boolean value) {
        retard = value;
    }


    public boolean 
    getRetard() {
        return(retard);
    }

    
    public void 
    reset(int x, int y, int xspeed, int yspeed) {
        this.x = x;
        this.y = y;
        startX = x;
        startY = y;

        this.XSpeed    = xspeed;
        this.YSpeed    = yspeed;
        originalXSpeed = xspeed;
        originalYSpeed = yspeed;
    }


    public void 
    reset(int x, int y, int xspeed, int yspeed, long startTime) {
        reset(x, y, xspeed, yspeed);

        xStartTime = startTime;
        yStartTime = startTime;
    }


    public boolean 
    hit(Racket racket) {
        int top    = racket.y;
        int bottom = racket.y + racket.height;

        int   xDiff = racket.x - x;
        float slope = YSpeed / XSpeed;
        int   tempY = y + Math.round(xDiff * slope);

        if ((tempY + diameter > top) && (tempY < bottom) ) {

// Put the ball back to the racket x position

            reset(racket.x, tempY, XSpeed, YSpeed, System.currentTimeMillis());
            return(true);
        } else {
            return(false);
        }
    }


    public void 
    reverse(Racket racket) {
        int top    = racket.y;
        int bottom = racket.y + racket.height;

        XSpeed = -XSpeed;
        int ballCenter = y + diameter / 2;
        int paddleCenter = top + (bottom - top) / 2;

        int fromCenter = paddleCenter - ballCenter;         

// positive, if ball in top half of paddle
// negative, if ball in bottom half of paddle

        if (YSpeed == 0) {
            YSpeed = 1;
	}

        float percentFromCenter = (float) fromCenter /
		                  (float) ((bottom - top) / 2);

        if (Ball_Debug) {
	    System.err.println("reverse() YSpeed = " + YSpeed + 
			       " percentFromCenter = " + percentFromCenter + 
			       " fromCenter = " + fromCenter +
              		       " halfpaddle = " + (bottom - top) / 2);
	}

/* Assert
 * Paddle: ||
 *         ||__ if ballCenter is here or above, then percentFromCenter > 0 (+ve)
 *         ||   if ballCenter is here or below, pFC < 0 (-ve)
 *         ||
 */

// If ball is heading down, want top of paddle to be negative and bottom 
// to be positive.

        if (YSpeed > 0) {
            if (percentFromCenter > 0) {
                percentFromCenter = -percentFromCenter;
	    }
        }

        if (percentFromCenter > 0) {      // if percentFromCenter is positive
            YSpeed = (int) (YSpeed * (1 + percentFromCenter));
        } else {
            YSpeed = (int) (YSpeed * (-1 + percentFromCenter));
	}

        if (Ball_Debug) {
	    System.err.println("reverse() YSpeed=" + YSpeed);
	}

        reset(x, y, XSpeed, YSpeed, System.currentTimeMillis());
    }


    public long expectedMessageTime = 0;


    public void 
    adjustSpeed(int distance, long netDelay) {

	if (Ball_Debug) {
            System.err.println("adjust()1: distance=" + distance + 
			       " netDelay=" + netDelay +
              		       " XSpeed=" + XSpeed + 
			       " originalXSpeed=" + originalXSpeed +
                               " YSpeed=" + YSpeed + 
			       " originalYSpeed=" + originalYSpeed);
	}

        XSpeed = originalXSpeed;
        YSpeed = originalYSpeed;

// travelTime, netDelayTime, and newTravelTime are in seconds

        float travelTime    = (float) ((float) distance) / (float) XSpeed;
        float netDelayTime  = (float) (netDelay / 1000.0) / 2;
        float newTravelTime = travelTime + netDelayTime;

        if (Ball_Debug) {
	    System.err.println("--- retard=" + retard +
              		       " adjustSpeed()1.5 travelTime=" + travelTime +
              		       " netDelayTime=" + netDelayTime +
              		       " newTravelTime =" + newTravelTime); 
	}

        long current = System.currentTimeMillis();
        expectedMessageTime = current +
		 Math.round(retard ? (newTravelTime*1000) : (travelTime));

        if (Ball_Debug) {
	    System.err.println("Current time: " + current +
              		" Expected reception time: " + expectedMessageTime);
	}

        if (retard) {
            XSpeed = Math.round(distance / newTravelTime);
            YSpeed = Math.round(YSpeed * (float) ((float) XSpeed /
                                                  (float) originalXSpeed));
        }

        if (Ball_Debug) {
	    System.err.println("adjust()2: distance=" + distance + 
			       " netDelay=" + netDelay +
              		       " XSpeed=" + XSpeed + 
			       " originalXSpeed=" + originalXSpeed +
              		       " YSpeed=" + YSpeed + 
			       " originalYSpeed=" + originalYSpeed + "\n");
	}
    }


    public void 
    paint(Graphics g) {
        g.setColor(Color.yellow);
        g.fillOval(x, y, diameter, diameter);
    }


    public String 
    toString() {
        return(super.toString() + " XSpeed=" + XSpeed + " YSpeed=" + YSpeed);
    }
}
